@@ -1,6 +1,6 @@
-/* graymilter - simple graylist mail filter module
+/* graymilter - simple laterlist mail filter module
 **
-** Copyright  2004 by Jef Poskanzer <jef@mail.acme.com>.
+** Copyright  2004,2025 by Jef Poskanzer <jef@mail.acme.com>.
 ** All rights reserved.
 **
 ** Redistribution and use in source and binary forms, with or without
@@ -27,34 +27,19 @@
 ** For commentary on this license please see http://www.acme.com/license.html
 */
 
-#ifdef STDC_HEADERS
-#include <stdlib.h>
-#include <string.h>
-#endif
-#ifdef HAVE_STDIO_H
-#include <stdio.h>
-#endif
-#ifdef HAVE_UNISTD_H
 #include <unistd.h>
-#endif
-#ifdef HAVE_SYSLOG_H
+#include <stdlib.h>
+#include <stdbool.h>
+#include <stdio.h>
+#include <string.h>
 #include <syslog.h>
-#endif
-#ifdef HAVE_ERRNO_H
 #include <errno.h>
-#endif
-#ifdef HAVE_SYS_TYPES_H
+#include <sysexits.h>
 #include <sys/types.h>
-#endif
-#ifdef HAVE_SYS_STAT_H
 #include <sys/stat.h>
-#endif
-#ifdef HAVE_PWD_H
 #include <pwd.h>
-#endif
-#ifdef HAVE_GRP_H
 #include <grp.h>
-#endif
+#include <pthread.h>
 #ifdef HAVE_SYS_SOCKET_H
 #include <sys/socket.h>
 #endif
@@ -64,14 +49,11 @@
 #ifdef HAVE_ARPA_INET_H
 #include <arpa/inet.h>
 #endif
-#ifdef HAVE_PTHREAD_H
-#include <pthread.h>
-#endif
 
-#include "libmilter/mfapi.h"
+#include <libmilter/mfapi.h>
 
 #include "version.h"
-#include "iparray.h"
+#include "iptab.h"
 
 
 /* Defines. */
@@ -79,8 +61,9 @@
 #define MIN_UPDATE_INTERVAL 10	/* min seconds between updates */
 #define MIN_FILE_AGE 30		/* min age of changed file before reading */
 
-#define DEFAULT_GRAYTIME 1500            /* 25 minutes */
-#define DEFAULT_WHITETIME 172800         /* two days */
+#define DEFAULT_LATERTIME 1500		/* 25 minutes */
+#define DEFAULT_ALLOWTIME 172800	/* two days */
+#define MAX_LISTS 10
 #define N_LISTS 10
 
 #define min(a,b) ((a) < (b) ? (a) : (b))
@@ -92,17 +75,17 @@
 static void usage( void );
 
 static void init_uid( const char* user );
-static void init_socket( const char* sockarg );
-static void fini_socket( const char* sockarg );
-static void unlink_socket( const char* sockarg );
+static void init_socket( char* sockarg );
+static void term_socket( char* sockarg );
 static void init_daemon( void );
 static void init_pidfile( const char* pidfilename );
-static void fini_pidfile( const char* pidfilename );
-static void init_iparrays( void );
-static void fini_iparrays( void );
+static void term_pidfile( const char* pidfilename );
+static void init_iptabs( void );
+static void term_iptabs( void );
 
-static int stat_file( time_t current_time, time_t mtime, char* filename );
-static time_t read_file( iparray list, char* filename );
+static void add_localhosts( iptab list );
+static bool stat_file( time_t current_time, time_t mtime, char* filename );
+static time_t read_file( iptab list, char* filename );
 static void trim( char* str );
 static void update( void );
 
@@ -117,18 +100,18 @@ static sfsistat gray_close( SMFICTX* ctx );
 /* Globals. */
 
 static char* argv0;
-static int graytime, whitetime;
-static char* initial_whitelist_filename;
-static time_t initial_whitelist_mtime;
-static iparray initial_whitelist;
-static octets* graylists[N_LISTS];
-static int max_graylists[N_LISTS];
-static int n_graylists[N_LISTS];
-static iparray whitelists[N_LISTS];
-static int n_whitelists[N_LISTS];
-static octets localhost = { 127, 0, 0, 1 };
-static int update_interval, gray_update_interval, white_update_interval;
-static time_t last_update, last_gray_update, last_white_update;
+static int latertime, allowtime;
+static char* initial_allowlist_filenames[MAX_LISTS];
+static time_t initial_allowlist_mtimes[MAX_LISTS];
+static iptab initial_allowlists[MAX_LISTS];
+static int n_initial_allowlists;
+static ipaddress* laterlists[N_LISTS];
+static int size_laterlists[N_LISTS];
+static int count_laterlists[N_LISTS];
+static iptab allowlists[N_LISTS];
+static int n_allowlists[N_LISTS];
+static int update_interval, later_update_interval, allow_update_interval;
+static time_t last_update, last_later_update, last_allow_update;
 static pthread_mutex_t lock;
 
 static struct smfiDesc smfilter =
@@ -145,7 +128,10 @@ static struct smfiDesc smfilter =
     NULL,		/* body block filter */
     NULL,		/* end of message */
     NULL,		/* message aborted */
-    gray_close		/* connection cleanup */
+    gray_close,		/* connection cleanup */
+    NULL,		/* unrecognized / unimplemented command */
+    NULL,		/* DATA command filter  */
+    NULL		/* negotiation  */
     };
 
 
@@ -165,30 +151,40 @@ main( int argc, char** argv )
     else
 	argv0 = argv[0];
 
+    openlog( argv0, LOG_PERROR, LOG_MAIL );
+
     /* Parse args. */
-    graytime = DEFAULT_GRAYTIME;
-    whitetime = DEFAULT_WHITETIME;
-    initial_whitelist_filename = (char*) 0;
+    latertime = DEFAULT_LATERTIME;
+    allowtime = DEFAULT_ALLOWTIME;
+    n_initial_allowlists = 1;	/* 0 is reserved for local addresses */
     user = (char*) 0;
     pidfilename = (char*) 0;
     nodaemon = 0;
     argn = 1;
     while ( argn < argc && argv[argn][0] == '-' && argv[argn][1] != '\0' )
 	{
-	if ( strncmp( argv[argn], "-graytime", strlen( argv[argn] ) ) == 0 && argn + 1 < argc )
+	if ( ( strncmp( argv[argn], "-latertime", strlen( argv[argn] ) ) == 0 ||
+	       strncmp( argv[argn], "-graytime", strlen( argv[argn] ) ) == 0 ) && argn + 1 < argc )
 	    {
 	    ++argn;
-	    graytime = atoi( argv[argn] );
+	    latertime = atoi( argv[argn] );
 	    }
-	else if ( strncmp( argv[argn], "-whitetime", strlen( argv[argn] ) ) == 0 && argn + 1 < argc )
+	else if ( ( strncmp( argv[argn], "-allowtime", strlen( argv[argn] ) ) == 0 ||
+	            strncmp( argv[argn], "-whitetime", strlen( argv[argn] ) ) == 0 ) && argn + 1 < argc )
 	    {
 	    ++argn;
-	    whitetime = atoi( argv[argn] );
+	    allowtime = atoi( argv[argn] );
 	    }
-	else if ( strncmp( argv[argn], "-initialwhitelist", max( strlen( argv[argn] ), 3 ) ) == 0 && argn + 1 < argc )
+	else if ( ( strncmp( argv[argn], "-initialallowlist", max( strlen( argv[argn] ), 3 ) ) == 0 ||
+	            strncmp( argv[argn], "-initialwhitelist", max( strlen( argv[argn] ), 3 ) ) == 0 ) && argn + 1 < argc )
 	    {
+	    if ( n_initial_allowlists == MAX_LISTS )
+		{
+		syslog( LOG_ERR, "too many initial allowlists" );
+		exit( EX_USAGE );
+		}
 	    ++argn;
-	    initial_whitelist_filename = argv[argn];
+	    initial_allowlist_filenames[n_initial_allowlists++] = argv[argn];
 	    }
 	else if ( strncmp( argv[argn], "-user", max( strlen( argv[argn] ), 3 ) ) == 0 && argn + 1 < argc )
 	    {
@@ -213,51 +209,52 @@ main( int argc, char** argv )
     sockarg = argv[argn++];
     if ( argn != argc )
 	usage();
-    if ( graytime < 1 || whitetime < 1 || whitetime < graytime )
+    if ( latertime < 1 || allowtime < 1 || allowtime < latertime )
 	usage();
 
-    openlog( argv0, 0, LOG_MAIL );
-    init_uid( user );
     init_socket( sockarg );
+    init_pidfile( pidfilename );
+    init_uid( user );
     if ( ! nodaemon )
 	init_daemon();
-    init_pidfile( pidfilename );
-    init_iparrays();
+    init_iptabs();
 
     if ( pthread_mutex_init( &lock, (pthread_mutexattr_t*) 0 ) != 0 )
 	{
-	(void) fprintf( stderr, "%s: pthread_mutex_init failed\n", argv0 );
-	exit( 1 );
+	syslog( LOG_ERR, "pthread_mutex_init() failed" );
+	exit( EX_OSERR );
 	}
 
-    gray_update_interval = graytime / N_LISTS;
-    white_update_interval = whitetime / N_LISTS;
-    if ( initial_whitelist_filename != (char*) 0 )
+    later_update_interval = latertime / N_LISTS;
+    allow_update_interval = allowtime / N_LISTS;
+    if ( n_initial_allowlists > 1 )
 	update_interval = MIN_UPDATE_INTERVAL;
     else
-	update_interval = min( gray_update_interval, white_update_interval ) / 2;
-    last_update = last_gray_update = last_white_update = time( (time_t*) 0 );
+	update_interval = min( later_update_interval, allow_update_interval ) / 2;
+    last_update = last_later_update = last_allow_update = time( (time_t*) 0 );
 
+    syslog( LOG_NOTICE, "%s %s starting", GRAYMILTER_PROGRAM, GRAYMILTER_VERSION );
     if ( smfi_main() == MI_FAILURE )
 	{
 	syslog( LOG_ERR, "smfi_main() failed" );
-	exit( 1 );
+	exit( EX_OSERR );
 	}
+    syslog( LOG_NOTICE, "%s %s terminating", GRAYMILTER_PROGRAM, GRAYMILTER_VERSION );
 
     (void) pthread_mutex_destroy( &lock );
-    fini_iparrays();
-    fini_pidfile( pidfilename );
-    fini_socket( sockarg );
+    term_iptabs();
+    term_pidfile( pidfilename );
+    term_socket( sockarg );
     closelog();
-    exit( 0 );
+    exit( EX_OK );
     }
 
 
 static void
 usage( void )
     {
-    (void) fprintf( stderr, "usage:  %s [-graytime seconds] [-whitetime seconds] [-initialwhitelist file] [-user user] [-pidfile filename] [-nodaemon|-X] socket\n", argv0 );
-    exit( 1 );
+    (void) fprintf( stderr, "usage:  %s [-latertime/-graytime seconds] [-allowtime/-whitetime seconds] [-initialallowlist/-initialwhitelist file] [-user user] [-pidfile filename] [-nodaemon|-X] socket\n", argv0 );
+    exit( EX_USAGE );
     }
 
 
@@ -270,9 +267,9 @@ init_uid( const char* user )
 
     if ( getuid() == 0 )
 	{
-	/* If we're root, the very first thing we do is become another user. */
+	/* If we're root, become another user. */
 	if ( user == (char*) 0 )
-	    (void) fprintf( stderr, "%s: warning: started as root but no --user flag specified\n", argv0 );
+	    syslog( LOG_WARNING, "warning: started as root but no -user flag specified" );
 	else
 	    {
 	    /* Is it a number? */
@@ -283,30 +280,30 @@ init_uid( const char* user )
 		pwd = getpwnam( user );
 	    if ( pwd == (struct passwd*) 0 )
 		{
-		(void) fprintf( stderr, "%s: unknown user: '%s'\n", argv0, user );
-		exit( 1 );
+		syslog( LOG_ERR, "unknown user: '%s'", user );
+		exit( EX_OSERR );
 		}
 	    /* Set aux groups to null. */
 	    if ( setgroups( 0, (gid_t*) 0 ) < 0 )
 		{
-		perror( "setgroups" );
-		exit( 1 );
+		syslog( LOG_ERR, "setgroups - %m" );
+		exit( EX_OSERR );
 		}
 	    /* Set primary group. */
 	    if ( setgid( pwd->pw_gid ) < 0 )
 		{
-		perror( "setgid" );
-		exit( 1 );
+		syslog( LOG_ERR, "setgid - %m" );
+		exit( EX_OSERR );
 		}
 
 	    /* Try setting aux groups correctly - not critical if this fails. */
 	    if ( initgroups( user, pwd->pw_gid ) < 0 )
-		    perror( "initgroups" );
+		syslog( LOG_WARNING, "initgroups - %m" );
 	    /* Set uid. */
 	    if ( setuid( pwd->pw_uid ) < 0 )
 		{
-		perror( "setuid" );
-		exit( 1 );
+		syslog( LOG_ERR, "setuid - %m" );
+		exit( EX_OSERR );
 		}
 	    }
 	}
@@ -315,63 +312,37 @@ init_uid( const char* user )
 	/* If we're not root but got a -user flag anyway, that's an error. */
 	if ( user != (char*) 0 )
 	    {
-	    (void) fprintf( stderr, "%s: can't switch users if not started as root\n", argv0 );
-	    exit( 1 );
+	    syslog( LOG_ERR, "can't switch users if not started as root" );
+	    exit( EX_USAGE );
 	    }
 	}
     }
 
 
 static void
-init_socket( const char* sockarg )
+init_socket( char* sockarg )
     {
-    unlink_socket( sockarg );
-
     /* Harden our umask so that the new socket gets created securely. */
     umask( 0077 );
 
     /* Initialize milter stuff. */
-    smfi_setconn( (char*) sockarg );
     if ( smfi_register( smfilter ) == MI_FAILURE )
 	{
-	(void) fprintf( stderr, "%s: smfi_register() failed\n", argv0 );
-	exit( 1 );
+	syslog( LOG_ERR, "smfi_register() failed" );
+	exit( EX_OSERR );
+	}
+    smfi_setconn( sockarg );
+    if ( smfi_opensocket( true ) == MI_FAILURE )
+	{
+	syslog( LOG_ERR, "smfi_opensocket() failed" );
+	exit( EX_OSERR );
 	}
     }
 
 
 static void
-fini_socket( const char* sockarg )
+term_socket( char* sockarg )
     {
-    unlink_socket( sockarg );
-    }
-
-
-static void
-unlink_socket( const char* sockarg )
-    {
-    /* Remove old local socket.  Actually we should probably stat
-    ** the file and check that it is in fact a socket before removing
-    ** it.
-    */
-    if ( strncasecmp( sockarg, "unix:", 5 ) == 0 )
-	{
-	if ( unlink( &sockarg[5] ) < 0 )
-	    if ( errno != ENOENT )
-		perror( &sockarg[5] );
-	}
-    else if ( strncasecmp( sockarg, "local:", 6 ) == 0 )
-	{
-	if ( unlink( &sockarg[6] ) < 0 )
-	    if ( errno != ENOENT )
-		perror( &sockarg[6] );
-	}
-    else if ( strchr( sockarg, ':' ) == (char*) 0 )
-	{
-	if ( unlink( sockarg ) < 0 )
-	    if ( errno != ENOENT )
-		perror( sockarg );
-	}
     }
 
 
@@ -382,8 +353,8 @@ init_daemon( void )
 #ifdef HAVE_DAEMON
     if ( daemon( 0, 0 ) < 0)
 	{
-	perror( "daemon" );
-	exit( 1 );
+	syslog( LOG_ERR, "daemon - %m" );
+	exit( EX_OSERR );
 	}
 #else /* HAVE_DAEMON */
     switch ( fork() )
@@ -391,11 +362,10 @@ init_daemon( void )
 	case 0:
 	    break;  
 	case -1:
-	    syslog( LOG_CRIT, "fork: %m" );
-	    perror( "fork" );
-	    exit( 1 );
+	    syslog( LOG_ERR, "fork - %m" );
+	    exit( EX_OSERR );
 	default:
-	    exit( 0 );
+	    exit( EX_OK );
 	}
 #ifdef HAVE_SETSID
     setsid();
@@ -424,7 +394,7 @@ init_pidfile( const char* pidfilename )
 
 
 static void
-fini_pidfile( const char* pidfilename )
+term_pidfile( const char* pidfilename )
     {
     if ( pidfilename != (char*) 0 )
 	(void) unlink( pidfilename );
@@ -432,89 +402,101 @@ fini_pidfile( const char* pidfilename )
 
 
 static void
-init_iparrays( void )
+init_iptabs( void )
     {       
     int i;
 
-    if ( initial_whitelist_filename != (char*) 0 )
+    initial_allowlists[0] = iptab_new();
+    add_localhosts( initial_allowlists[0] );
+    for ( i = 1; i < n_initial_allowlists; ++i )
 	{
-	initial_whitelist = iparray_new();
-	if ( initial_whitelist == (iparray) 0 )
+	initial_allowlists[i] = iptab_new();
+	if ( initial_allowlists[i] == (iptab) 0 )
 	    {
-	    (void) fprintf( stderr, "%s: whitelist create failed\n", argv0 );
-	    exit( 1 );
+	    syslog( LOG_ERR, "allowlist create failed" );
+	    exit( EX_OSERR );
 	    }
-	initial_whitelist_mtime = read_file( initial_whitelist, initial_whitelist_filename );
+	initial_allowlist_mtimes[i] = read_file( initial_allowlists[i], initial_allowlist_filenames[i] );
 	}
 
     for ( i = 0; i < N_LISTS; ++i )
 	{
-	max_graylists[i] = 1000;	/* arbitrary */
-	graylists[i] = (octets*) malloc( max_graylists[i] * sizeof(octets) );
-	if ( graylists[i] == (octets*) 0 )
+	size_laterlists[i] = 1000;	/* arbitrary */
+	laterlists[i] = (ipaddress*) malloc( size_laterlists[i] * sizeof(ipaddress) );
+	if ( laterlists[i] == (ipaddress*) 0 )
 	    {
-	    (void) fprintf( stderr, "%s: graylist create failed\n", argv0 );
-	    exit( 1 );
+	    syslog( LOG_ERR, "laterlist create failed" );
+	    exit( EX_OSERR );
 	    }
-	n_graylists[i] = 0;
-	whitelists[i] = iparray_new();
-	if ( whitelists[i] == (iparray) 0 )
+	count_laterlists[i] = 0;
+	allowlists[i] = iptab_new();
+	if ( allowlists[i] == (iptab) 0 )
 	    {
-	    (void) fprintf( stderr, "%s: whitelist create failed\n", argv0 );
-	    exit( 1 );
+	    syslog( LOG_ERR, "allowlist create failed" );
+	    exit( EX_OSERR );
 	    }
-	n_whitelists[i] = 0;
+	n_allowlists[i] = 0;
 	}
     }
 
 
 static void
-fini_iparrays( void )
+term_iptabs( void )
     {
     int i;
 
     for ( i = 0; i < N_LISTS; ++i )
 	{
-	free( (void*) graylists[i] );
-	iparray_delete( whitelists[i] );
+	free( (void*) laterlists[i] );
+	iptab_delete( allowlists[i] );
 	}
-    if ( initial_whitelist_filename != (char*) 0 )
-	iparray_delete( initial_whitelist );
-    iparray_fini();
+    for ( i = 0; i < n_initial_allowlists; ++i )
+	iptab_delete( initial_allowlists[i] );
     }
 
 
-/* Returns 1 if the file is still current, else 0. */
-static int
+static void
+add_localhosts( iptab list )
+    {
+    const ipaddress v4localhost = { { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0xff, 0xff, 127, 0, 0, 0 }, 104 };
+    const ipaddress v6localhost = { { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1 }, 128 };
+
+    (void) iptab_add( list, &v4localhost );
+    (void) iptab_add( list, &v6localhost );
+    }
+
+
+/* Returns true if the file is still current, else false. */
+static bool
 stat_file( time_t current_time, time_t mtime, char* filename )
     {
     struct stat sb;
 
     if ( stat( filename, &sb ) < 0 )
-	return 1;	/* Can't stat it?  Ignore. */
+	return true;	/* Can't stat it?  Ignore. */
     if ( sb.st_mtime == mtime )
-	return 1;	/* Unchanged. */
+	return true;	/* Unchanged. */
     if ( current_time - sb.st_mtime < MIN_FILE_AGE )
-	return 1;	/* Not old enough, we'll catch it next time. */
-    return 0;		/* Changed. */
+	return true;	/* Not old enough, we'll catch it next time. */
+    return false;		/* Changed. */
     }
 
 
 /* Reads one file into the database. */
 static time_t
-read_file( iparray list, char* filename )
+read_file( iptab list, char* filename )
     {
     FILE* fp;
     struct stat sb;
     time_t mtime;
-    char line[10000];
-    octets ip;
+    char line[1000];
+    ipaddress ipa;
 
     syslog( LOG_INFO, "reading %s", filename );
     fp = fopen( filename, "r" );
     if ( fp == (FILE*) 0 )
 	{
-	syslog( LOG_ERR, "%s - %m", filename );
+	syslog( LOG_ERR, "fopen '%s' - %m", filename );
 	mtime = (time_t) -1;
 	}
     else
@@ -528,8 +510,8 @@ read_file( iparray list, char* filename )
 	    trim( line );
 	    if ( line[0] == '\0' )
 		continue;
-	    if ( iparray_parse_octets( line, &ip ) )
-		(void) iparray_incr( list, ip );
+	    if ( iptab_parse_address( line, &ipa ) )
+		(void) iptab_add( list, &ipa );
 	    else
 		syslog( LOG_INFO, "unparsable IP address - \"%s\"", line );
 	    }
@@ -562,9 +544,9 @@ update( void )
     {
     time_t current_time;
     int i;
-    octets* tmp_graylist;
-    int tmp_max_graylist;
-    iparray tmp_whitelist;
+    ipaddress* tmp_laterlist;
+    int tmp_size_laterlist;
+    iptab tmp_allowlist;
 
     current_time = time( (time_t*) 0 );
     if ( current_time - last_update < update_interval )
@@ -573,54 +555,56 @@ update( void )
 
     if ( pthread_mutex_lock( &lock ) == 0 )
 	{
-	if ( initial_whitelist_filename != (char*) 0 )
-	    if ( ! stat_file( current_time, initial_whitelist_mtime, initial_whitelist_filename ) )
-		{
-		syslog( LOG_INFO, "initial whitelist file changed - autoupdating" );
-		iparray_clear( initial_whitelist );
-		initial_whitelist_mtime = read_file( initial_whitelist, initial_whitelist_filename );
-		}
-
-	while ( current_time - last_gray_update >= gray_update_interval )
+	for ( i = 1; i < n_initial_allowlists; ++i )
 	    {
-	    /* Graduate addresses from the last graylist to the first whitelist. */
-	    syslog( LOG_INFO, "graduating %d addresses to whitelist", n_graylists[N_LISTS-1] );
-	    for ( i = 0; i < n_graylists[N_LISTS-1]; ++i )
-		(void) iparray_incr( whitelists[0], graylists[N_LISTS-1][i] );
-	    n_whitelists[0] += n_graylists[N_LISTS-1];
-	    /* Rotate the graylists. */
-	    tmp_graylist = graylists[N_LISTS-1];
-	    tmp_max_graylist = max_graylists[N_LISTS-1];
-	    for ( i = N_LISTS - 1; i > 0; --i )
+	    if ( ! stat_file( current_time, initial_allowlist_mtimes[i], initial_allowlist_filenames[i] ) )
 		{
-		graylists[i] = graylists[i-1];
-		max_graylists[i] = max_graylists[i-1];
-		n_graylists[i] = n_graylists[i-1];
+		syslog( LOG_INFO, "initial allowlist file %d changed - autoupdating", i );
+		iptab_clear( initial_allowlists[i] );
+		initial_allowlist_mtimes[i] = read_file( initial_allowlists[i], initial_allowlist_filenames[i] );
 		}
-	    graylists[0] = tmp_graylist;
-	    max_graylists[0] = tmp_max_graylist;
-	    /* Clear out the first graylist. */
-	    n_graylists[0] = 0;
-	    /* Bump update time. */
-	    last_gray_update += gray_update_interval;
 	    }
 
-	while ( current_time - last_white_update >= white_update_interval )
+	while ( current_time - last_later_update >= later_update_interval )
 	    {
-	    syslog( LOG_INFO, "expiring %d addresses from whitelist", n_whitelists[N_LISTS-1] );
-	    /* Rotate the whitelists. */
-	    tmp_whitelist = whitelists[N_LISTS-1];
+	    /* Graduate addresses from the last laterlist to the first allowlist. */
+	    syslog( LOG_INFO, "graduating %d addresses to allowlist", count_laterlists[N_LISTS-1] );
+	    for ( i = 0; i < count_laterlists[N_LISTS-1]; ++i )
+		(void) iptab_add( allowlists[0], &laterlists[N_LISTS-1][i] );
+	    n_allowlists[0] += count_laterlists[N_LISTS-1];
+	    /* Rotate the laterlists. */
+	    tmp_laterlist = laterlists[N_LISTS-1];
+	    tmp_size_laterlist = size_laterlists[N_LISTS-1];
 	    for ( i = N_LISTS - 1; i > 0; --i )
 		{
-		whitelists[i] = whitelists[i-1];
-		n_whitelists[i] = n_whitelists[i-1];
+		laterlists[i] = laterlists[i-1];
+		size_laterlists[i] = size_laterlists[i-1];
+		count_laterlists[i] = count_laterlists[i-1];
 		}
-	    whitelists[0] = tmp_whitelist;
-	    /* Clear out the first whitelist. */
-	    iparray_clear( whitelists[0] );
-	    n_whitelists[0] = 0;
+	    laterlists[0] = tmp_laterlist;
+	    size_laterlists[0] = tmp_size_laterlist;
+	    /* Clear out the first laterlist. */
+	    count_laterlists[0] = 0;
 	    /* Bump update time. */
-	    last_white_update += white_update_interval;
+	    last_later_update += later_update_interval;
+	    }
+
+	while ( current_time - last_allow_update >= allow_update_interval )
+	    {
+	    syslog( LOG_INFO, "expiring %d addresses from allowlist", n_allowlists[N_LISTS-1] );
+	    /* Rotate the allowlists. */
+	    tmp_allowlist = allowlists[N_LISTS-1];
+	    for ( i = N_LISTS - 1; i > 0; --i )
+		{
+		allowlists[i] = allowlists[i-1];
+		n_allowlists[i] = n_allowlists[i-1];
+		}
+	    allowlists[0] = tmp_allowlist;
+	    /* Clear out the first allowlist. */
+	    iptab_clear( allowlists[0] );
+	    n_allowlists[0] = 0;
+	    /* Bump update time. */
+	    last_allow_update += allow_update_interval;
 	    }
 
 	(void) pthread_mutex_unlock( &lock );
@@ -630,7 +614,7 @@ update( void )
 
 /* The private data struct. */
 struct connection_data {
-    octets ip;
+    ipaddress ipa;
     int action;
     char *connhost;
     };
@@ -650,67 +634,105 @@ static sfsistat
 gray_connect( SMFICTX* ctx, char* connhost, _SOCK_ADDR* connaddr )
     {
     struct connection_data* cd;
-    struct sockaddr_in* sin;
-    char* char_addr;
     int i;
+    char str[100];
 
     update();
 
     if ( connaddr == (_SOCK_ADDR*) 0 )
-	return SMFIS_ACCEPT;	/* can't deal with it */
-
-    cd = (struct connection_data*) malloc( sizeof(struct connection_data) );
-    if ( cd == (struct connection_data*) 0 )
-	return SMFIS_TEMPFAIL;
-    (void) smfi_setpriv( ctx, (void*) cd );
-    cd->action = ACTION_UNKNOWN;
-
-    sin = (struct sockaddr_in*) connaddr;
-
-    char_addr = (char*) &sin->sin_addr.s_addr;
-    cd->ip.a = char_addr[0];
-    cd->ip.b = char_addr[1];
-    cd->ip.c = char_addr[2];
-    cd->ip.d = char_addr[3];
-    cd->ip.w = 32;
-
-    cd->connhost = strdup( connhost );	/* don't care about failure */
-
-    /* Add the address to the first graylist. */
-    if ( n_graylists[0] >= max_graylists[0] )
 	{
-	/* Oosp, got to expand the list. */
-	if ( pthread_mutex_lock( &lock ) != 0 )
-	    return SMFIS_TEMPFAIL;
-	max_graylists[0] *= 2;
-	syslog( LOG_NOTICE, "expanding a graylist to %d items", max_graylists[0] );
-	graylists[0] = (octets*) realloc( (void*) graylists[0], max_graylists[0] * sizeof(octets) );
-	(void) pthread_mutex_unlock( &lock );
-	}
-    graylists[0][n_graylists[0]] = cd->ip;
-    ++n_graylists[0];
-
-    /* Special case for localhost. */
-    if ( cd->ip.a == localhost.a )	/* localhost is actually a class-A network */
-	{
-	syslog( LOG_INFO, "%d.%d.%d.%d is localhost - accepting", cd->ip.a, cd->ip.b, cd->ip.c, cd->ip.d );
+	syslog( LOG_INFO, "no sockaddr - accepting" );
 	return SMFIS_ACCEPT;
 	}
 
-    /* Check if the address is in the optional initial whitelist. */
-    if ( initial_whitelist_filename != (char*) 0 )
-	if ( iparray_get( initial_whitelist, cd->ip ) > 0 )
+    cd = (struct connection_data*) malloc( sizeof(struct connection_data) );
+    if ( cd == (struct connection_data*) 0 )
+	{
+	syslog( LOG_ERR, "couldn't allocate connection_data" );
+	return SMFIS_ACCEPT;
+	}
+    (void) smfi_setpriv( ctx, (void*) cd );
+    cd->action = ACTION_UNKNOWN;
+
+    if ( connaddr->sa_family == AF_INET )
+	{
+	struct sockaddr_in* sa_in;
+	unsigned char* uchar_addr;
+
+	sa_in = (struct sockaddr_in*) ( (void*) connaddr );	/* extra cast to avoid alignment warning */
+	uchar_addr = (unsigned char*) &sa_in->sin_addr.s_addr;
+	cd->ipa.octets[0] = cd->ipa.octets[1] = cd->ipa.octets[2] = cd->ipa.octets[3] = cd->ipa.octets[4] = cd->ipa.octets[5] = cd->ipa.octets[6] = cd->ipa.octets[7] = cd->ipa.octets[8] = cd->ipa.octets[9] = 0;
+	cd->ipa.octets[10] = cd->ipa.octets[11] = 0xff;
+	cd->ipa.octets[12] = uchar_addr[0];
+	cd->ipa.octets[13] = uchar_addr[1];
+	cd->ipa.octets[14] = uchar_addr[2];
+	cd->ipa.octets[15] = uchar_addr[3];
+	cd->ipa.prefixlen = 128;
+	}
+    else if ( connaddr->sa_family == AF_INET6 )
+	{
+	struct sockaddr_in6* sa_in6;
+	unsigned char* uchar_addr;
+
+	sa_in6 = (struct sockaddr_in6*) ( (void*) connaddr );	/* extra cast to avoid alignment warning */
+	uchar_addr = (unsigned char*) &sa_in6->sin6_addr.s6_addr;
+	cd->ipa.octets[0] = uchar_addr[0];
+	cd->ipa.octets[1] = uchar_addr[1];
+	cd->ipa.octets[2] = uchar_addr[2];
+	cd->ipa.octets[3] = uchar_addr[3];
+	cd->ipa.octets[4] = uchar_addr[4];
+	cd->ipa.octets[5] = uchar_addr[5];
+	cd->ipa.octets[6] = uchar_addr[6];
+	cd->ipa.octets[7] = uchar_addr[7];
+	cd->ipa.octets[8] = uchar_addr[8];
+	cd->ipa.octets[9] = uchar_addr[9];
+	cd->ipa.octets[10] = uchar_addr[10];
+	cd->ipa.octets[11] = uchar_addr[11];
+	cd->ipa.octets[12] = uchar_addr[12];
+	cd->ipa.octets[13] = uchar_addr[13];
+	cd->ipa.octets[14] = uchar_addr[14];
+	cd->ipa.octets[15] = uchar_addr[15];
+	cd->ipa.prefixlen = 128;
+	}
+    else
+	{
+	syslog( LOG_INFO, "unknown address family - accepting" );
+	return SMFIS_ACCEPT;
+	}
+
+    cd->connhost = strdup( connhost );	/* don't care about failure */
+
+    /* Add the address to the first laterlist. */
+    if ( count_laterlists[0] >= size_laterlists[0] )
+	{
+	/* Oosp, got to expand the list. */
+	if ( pthread_mutex_lock( &lock ) != 0 )
 	    {
-	    syslog( LOG_INFO, "%d.%d.%d.%d (%s) is whitelisted - accepting", cd->ip.a, cd->ip.b, cd->ip.c, cd->ip.d, cd->connhost );
+	    syslog( LOG_ERR, "pthread_mutex_lock() failed" );
+	    return SMFIS_ACCEPT;
+	    }
+	size_laterlists[0] *= 2;
+	syslog( LOG_NOTICE, "expanding a laterlist to %d items", size_laterlists[0] );
+	laterlists[0] = (ipaddress*) realloc( (void*) laterlists[0], size_laterlists[0] * sizeof(ipaddress) );
+	(void) pthread_mutex_unlock( &lock );
+	}
+    laterlists[0][count_laterlists[0]] = cd->ipa;
+    ++count_laterlists[0];
+
+    /* Check if the address is in the initial allowlists. */
+    for ( i = 0; i < n_initial_allowlists; ++i )
+	if ( iptab_check( initial_allowlists[i], &cd->ipa ) )
+	    {
+	    syslog( LOG_INFO, "%s (%s) is allowlisted - accepting", iptab_format_address( &cd->ipa, str, sizeof(str) ), cd->connhost );
 	    return SMFIS_ACCEPT;
 	    }
 
-    /* Check if the address is in any of the whitelists. */
+    /* Check if the address is in any of the allowlists. */
     for ( i = 0; i < N_LISTS; ++i )
 	{
-	if ( iparray_get( whitelists[i], cd->ip ) > 0 )
+	if ( iptab_check( allowlists[i], &cd->ipa ) )
 	    {
-	    syslog( LOG_INFO, "%d.%d.%d.%d (%s) has been seen before - accepting", cd->ip.a, cd->ip.b, cd->ip.c, cd->ip.d, cd->connhost );
+	    syslog( LOG_INFO, "%s (%s) has been seen before - accepting", iptab_format_address( &cd->ipa, str, sizeof(str) ), cd->connhost );
 	    return SMFIS_ACCEPT;
 	    }
 	}
@@ -738,6 +760,7 @@ gray_envfrom( SMFICTX* ctx, char** fromargs )
     struct connection_data* cd = (struct connection_data*) smfi_getpriv( ctx );
     char* auth_type;
     char* verify;
+    char str[100];
 
     /* The temporary failure response can't happen in the connect handler
     ** because we haven't started speaking SMTP protocol yet.  You need to
@@ -755,15 +778,17 @@ gray_envfrom( SMFICTX* ctx, char** fromargs )
     if ( auth_type != (char*) 0 ||
          ( verify != (char*) 0 && strcmp( verify, "OK" ) == 0 ) )
 	{
-	syslog( LOG_INFO, "%d.%d.%d.%d (%s) is authenticated - accepting", cd->ip.a, cd->ip.b, cd->ip.c, cd->ip.d, cd->connhost );
+	syslog( LOG_INFO, "%s (%s) is authenticated - accepting", iptab_format_address( &cd->ipa, str, sizeof(str) ), cd->connhost );
 	return SMFIS_ACCEPT;
 	}
 
     if ( cd->action == ACTION_TEMPFAIL )
 	{
 	/* Send back the temporary failure code. */
-	/* syslog( LOG_INFO, "%d.%d.%d.%d (%s) is new - tempfailing", cd->ip.a, cd->ip.b, cd->ip.c, cd->ip.d, cd->connhost ); */
-	(void) smfi_setreply( ctx, "421", "4.3.2", "graylisted - please try again later" );
+#ifdef notdef
+	syslog( LOG_INFO, "%d.%d.%d.%d (%s) is new - tempfailing", cd->ipa.a, cd->ipa.b, cd->ipa.c, cd->ipa.d, cd->connhost );
+#endif
+	(void) smfi_setreply( ctx, "421", "4.3.2", "laterlisted - please try again later" );
 	return SMFIS_TEMPFAIL;
 	}
 
